@@ -27,6 +27,7 @@
 #include <sys/ipc.h>
 #endif
 
+#include "glib-compat.h"
 #include "spice-client.h"
 #include "spice-common.h"
 
@@ -70,6 +71,8 @@ struct _SpiceDisplayChannelPrivate {
     int                         nstreams;
     gboolean                    mark;
     guint                       mark_false_event_id;
+    GArray                      *monitors;
+    guint                       monitors_max;
 #ifdef WIN32
     HDC dc;
 #endif
@@ -82,6 +85,8 @@ enum {
     PROP_0,
     PROP_WIDTH,
     PROP_HEIGHT,
+    PROP_MONITORS,
+    PROP_MONITORS_MAX
 };
 
 enum {
@@ -100,9 +105,10 @@ static void spice_display_channel_up(SpiceChannel *channel);
 
 static void clear_surfaces(SpiceChannel *channel, gboolean keep_primary);
 static void clear_streams(SpiceChannel *channel);
-static display_surface *find_surface(SpiceDisplayChannelPrivate *c, int surface_id);
+static display_surface *find_surface(SpiceDisplayChannelPrivate *c, guint32 surface_id);
 static gboolean display_stream_render(display_stream *st);
 static void spice_display_channel_reset(SpiceChannel *channel, gboolean migrating);
+static void spice_display_channel_reset_capabilities(SpiceChannel *channel);
 static void destroy_canvas(display_surface *surface);
 
 /* ------------------------------------------------------------------ */
@@ -122,6 +128,9 @@ static void spice_display_channel_dispose(GObject *object)
 
 static void spice_display_channel_finalize(GObject *object)
 {
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(object)->priv;
+
+    g_clear_pointer(&c->monitors, g_array_unref);
     clear_surfaces(SPICE_CHANNEL(object), FALSE);
     clear_streams(SPICE_CHANNEL(object));
 
@@ -141,6 +150,8 @@ static void spice_display_channel_constructed(GObject *object)
     g_return_if_fail(c->images != NULL);
     g_return_if_fail(c->palettes != NULL);
 
+    c->monitors = g_array_new(FALSE, TRUE, sizeof(SpiceDisplayMonitorConfig));
+
     if (G_OBJECT_CLASS(spice_display_channel_parent_class)->constructed)
         G_OBJECT_CLASS(spice_display_channel_parent_class)->constructed(object);
 }
@@ -164,6 +175,14 @@ static void spice_display_get_property(GObject    *object,
         g_value_set_uint(value, surface ? surface->height : 0);
         break;
     }
+    case PROP_MONITORS: {
+        g_value_set_boxed(value, c->monitors);
+        break;
+    }
+    case PROP_MONITORS_MAX: {
+        g_value_set_uint(value, c->monitors_max);
+        break;
+    }
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
         break;
@@ -206,6 +225,7 @@ static void spice_display_channel_class_init(SpiceDisplayChannelClass *klass)
     channel_class->handle_msg   = spice_display_handle_msg;
     channel_class->channel_up   = spice_display_channel_up;
     channel_class->channel_reset = spice_display_channel_reset;
+    channel_class->channel_reset_capabilities = spice_display_channel_reset_capabilities;
 
     g_object_class_install_property
         (gobject_class, PROP_HEIGHT,
@@ -225,6 +245,40 @@ static void spice_display_channel_class_init(SpiceDisplayChannelClass *klass)
                            G_PARAM_READABLE |
                            G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceDisplayChannel:monitors:
+     *
+     * Current monitors configuration.
+     *
+     * Since: 0.13
+     */
+    g_object_class_install_property
+        (gobject_class, PROP_MONITORS,
+         g_param_spec_boxed("monitors",
+                            "Display monitors",
+                            "The monitors configuration",
+                            G_TYPE_ARRAY,
+                            G_PARAM_READABLE |
+                            G_PARAM_STATIC_STRINGS));
+
+    /**
+     * SpiceDisplayChannel:monitors-max:
+     *
+     * The maximum number of monitors the server or guest supports.
+     * May change during client lifetime, for instance guest may
+     * reboot or dynamically adjust this.
+     *
+     * Since: 0.13
+     */
+    g_object_class_install_property
+        (gobject_class, PROP_MONITORS_MAX,
+         g_param_spec_uint("monitors-max",
+                           "Max display monitors",
+                           "The current maximum number of monitors",
+                           1, G_MAXINT16, 1,
+                           G_PARAM_READABLE |
+                           G_PARAM_STATIC_STRINGS));
+
     /**
      * SpiceDisplayChannel::display-primary-create:
      * @display: the #SpiceDisplayChannel that emitted the signal
@@ -322,6 +376,43 @@ static void spice_display_channel_class_init(SpiceDisplayChannelClass *klass)
     rop3_init();
 }
 
+/**
+ * spice_display_get_primary:
+ * @channel:
+ * @surface_id:
+ * @primary:
+ *
+ * Retrieve primary display surface @surface_id.
+ *
+ * Returns: %TRUE if the primary surface was found and its details
+ * collected in @primary.
+ */
+gboolean spice_display_get_primary(SpiceChannel *channel, guint32 surface_id,
+                                   SpiceDisplayPrimary *primary)
+{
+    g_return_val_if_fail(SPICE_IS_DISPLAY_CHANNEL(channel), FALSE);
+    g_return_val_if_fail(primary != NULL, FALSE);
+
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
+    display_surface *surface = find_surface(c, surface_id);
+
+    if (surface == NULL)
+        return FALSE;
+
+    g_return_val_if_fail(surface->primary, FALSE);
+
+    primary->format = surface->format;
+    primary->width = surface->width;
+    primary->height = surface->height;
+    primary->stride = surface->stride;
+    primary->shmid = surface->shmid;
+    primary->data = surface->data;
+    primary->marked = c->mark;
+    SPICE_DEBUG("get primary %p", primary->data);
+
+    return TRUE;
+}
+
 /* signal trampoline---------------------------------------------------------- */
 
 struct SPICE_DISPLAY_PRIMARY_CREATE {
@@ -584,6 +675,12 @@ static HDC create_compatible_dc(void)
 }
 #endif
 
+static void spice_display_channel_reset_capabilities(SpiceChannel *channel)
+{
+    spice_channel_set_capability(SPICE_CHANNEL(channel), SPICE_DISPLAY_CAP_SIZED_STREAM);
+    spice_channel_set_capability(SPICE_CHANNEL(channel), SPICE_DISPLAY_CAP_MONITORS_CONFIG);
+}
+
 static void spice_display_channel_init(SpiceDisplayChannel *channel)
 {
     SpiceDisplayChannelPrivate *c;
@@ -597,6 +694,7 @@ static void spice_display_channel_init(SpiceDisplayChannel *channel)
 #if defined(WIN32)
     c->dc = create_compatible_dc();
 #endif
+    spice_display_channel_reset_capabilities(SPICE_CHANNEL(channel));
 }
 
 /* ------------------------------------------------------------------ */
@@ -670,10 +768,21 @@ static int create_canvas(SpiceChannel *channel, display_surface *surface)
     g_return_val_if_fail(surface->canvas != NULL, 0);
     ring_add(&c->surfaces, &surface->link);
 
-    if (surface->primary)
+    if (surface->primary) {
         emit_main_context(channel, SPICE_DISPLAY_PRIMARY_CREATE,
                           surface->format, surface->width, surface->height,
                           surface->stride, surface->shmid, surface->data);
+
+        if (!spice_channel_test_capability(channel, SPICE_DISPLAY_CAP_MONITORS_CONFIG)) {
+            g_array_set_size(c->monitors, 1);
+            SpiceDisplayMonitorConfig *config = &g_array_index(c->monitors, SpiceDisplayMonitorConfig, 0);
+            config->x = config->y = 0;
+            config->width = surface->width;
+            config->height = surface->height;
+            g_object_notify_main_context(G_OBJECT(channel), "monitors");
+        }
+    }
+
     return 0;
 }
 
@@ -702,7 +811,7 @@ static void destroy_canvas(display_surface *surface)
     surface->canvas = NULL;
 }
 
-static display_surface *find_surface(SpiceDisplayChannelPrivate *c, int surface_id)
+static display_surface *find_surface(SpiceDisplayChannelPrivate *c, guint32 surface_id)
 {
     display_surface *surface;
     RingItem *item;
@@ -962,7 +1071,7 @@ static void display_handle_stream_create(SpiceChannel *channel, SpiceMsgIn *in)
 static gboolean display_stream_schedule(display_stream *st)
 {
     guint32 time, d;
-    SpiceMsgDisplayStreamData *op;
+    SpiceStreamDataHeader *op;
     SpiceMsgIn *in;
 
     if (st->timeout)
@@ -989,6 +1098,72 @@ static gboolean display_stream_schedule(display_stream *st)
     return FALSE;
 }
 
+static SpiceRect *stream_get_dest(display_stream *st)
+{
+    if (st->msg_data == NULL ||
+        spice_msg_in_type(st->msg_data) != SPICE_MSG_DISPLAY_STREAM_DATA_SIZED) {
+        SpiceMsgDisplayStreamCreate *info = spice_msg_in_parsed(st->msg_create);
+
+        return &info->dest;
+    } else {
+        SpiceMsgDisplayStreamDataSized *op = spice_msg_in_parsed(st->msg_data);
+
+        return &op->dest;
+   }
+
+}
+
+static uint32_t stream_get_flags(display_stream *st)
+{
+    SpiceMsgDisplayStreamCreate *info = spice_msg_in_parsed(st->msg_create);
+
+    return info->flags;
+}
+
+G_GNUC_INTERNAL
+uint32_t stream_get_current_frame(display_stream *st, uint8_t **data)
+{
+    if (st->msg_data == NULL) {
+        *data = NULL;
+        return 0;
+    }
+
+    if (spice_msg_in_type(st->msg_data) == SPICE_MSG_DISPLAY_STREAM_DATA) {
+        SpiceMsgDisplayStreamData *op = spice_msg_in_parsed(st->msg_data);
+
+        *data = op->data;
+        return op->data_size;
+    } else {
+        SpiceMsgDisplayStreamDataSized *op = spice_msg_in_parsed(st->msg_data);
+
+        g_return_val_if_fail(spice_msg_in_type(st->msg_data) ==
+                             SPICE_MSG_DISPLAY_STREAM_DATA_SIZED, 0);
+        *data = op->data;
+        return op->data_size;
+   }
+
+}
+
+G_GNUC_INTERNAL
+void stream_get_dimensions(display_stream *st, int *width, int *height)
+{
+    g_return_if_fail(width != NULL);
+    g_return_if_fail(height != NULL);
+
+    if (st->msg_data == NULL ||
+        spice_msg_in_type(st->msg_data) != SPICE_MSG_DISPLAY_STREAM_DATA_SIZED) {
+        SpiceMsgDisplayStreamCreate *info = spice_msg_in_parsed(st->msg_create);
+
+        *width = info->stream_width;
+        *height = info->stream_height;
+    } else {
+        SpiceMsgDisplayStreamDataSized *op = spice_msg_in_parsed(st->msg_data);
+
+        *width = op->width;
+        *height = op->height;
+   }
+}
+
 /* main context */
 static gboolean display_stream_render(display_stream *st)
 {
@@ -1008,14 +1183,19 @@ static gboolean display_stream_render(display_stream *st)
         }
 
         if (st->out_frame) {
-            SpiceMsgDisplayStreamCreate *info = spice_msg_in_parsed(st->msg_create);
+            int width;
+            int height;
+            SpiceRect *dest;
             uint8_t *data;
             int stride;
 
+            stream_get_dimensions(st, &width, &height);
+            dest = stream_get_dest(st);
+
             data = st->out_frame;
-            stride = info->stream_width * sizeof(uint32_t);
-            if (!(info->flags & SPICE_STREAM_FLAGS_TOP_DOWN)) {
-                data += stride * (info->src_height - 1);
+            stride = width * sizeof(uint32_t);
+            if (!(stream_get_flags(st) & SPICE_STREAM_FLAGS_TOP_DOWN)) {
+                data += stride * (height - 1);
                 stride = -stride;
             }
 
@@ -1024,15 +1204,15 @@ static gboolean display_stream_render(display_stream *st)
 #ifdef WIN32
                 SPICE_DISPLAY_CHANNEL(st->channel)->priv->dc,
 #endif
-                &info->dest, data,
-                info->src_width, info->src_height, stride,
+                dest, data,
+                width, height, stride,
                 st->have_region ? &st->region : NULL);
 
             if (st->surface->primary)
                 g_signal_emit(st->channel, signals[SPICE_DISPLAY_INVALIDATE], 0,
-                    info->dest.left, info->dest.top,
-                    info->dest.right - info->dest.left,
-                    info->dest.bottom - info->dest.top);
+                    dest->left, dest->top,
+                    dest->right - dest->left,
+                    dest->bottom - dest->top);
         }
 
         st->msg_data = NULL;
@@ -1053,12 +1233,16 @@ static gboolean display_stream_render(display_stream *st)
 static void display_handle_stream_data(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
-    SpiceMsgDisplayStreamData *op = spice_msg_in_parsed(in);
+    SpiceStreamDataHeader *op = spice_msg_in_parsed(in);
     display_stream *st = c->streams[op->id];
     guint32 mmtime;
 
     mmtime = spice_session_get_mm_time(spice_channel_get_session(channel));
 
+    if (spice_msg_in_type(in) == SPICE_MSG_DISPLAY_STREAM_DATA_SIZED) {
+        SPICE_DEBUG("stream %d contains sized data", op->id);
+    }
+
     if (op->multi_media_time == 0) {
         g_critical("Received frame with invalid 0 timestamp! perhaps wrong graphic driver?");
         op->multi_media_time = mmtime + 100; /* workaround... */
@@ -1309,6 +1493,38 @@ static void display_handle_surface_destroy(SpiceChannel *channel, SpiceMsgIn *in
     free(surface);
 }
 
+/* coroutine context */
+static void display_handle_monitors_config(SpiceChannel *channel, SpiceMsgIn *in)
+{
+    SpiceMsgDisplayMonitorsConfig *config = spice_msg_in_parsed(in);
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
+    guint i;
+
+    g_return_if_fail(config != NULL);
+    g_return_if_fail(config->count > 0);
+
+    SPICE_DEBUG("monitors config: n: %d/%d", config->count, config->max_allowed);
+
+    c->monitors_max = config->max_allowed;
+    c->monitors = g_array_set_size(c->monitors, config->count);
+
+    for (i = 0; i < config->count; i++) {
+        SpiceDisplayMonitorConfig *mc = &g_array_index(c->monitors, SpiceDisplayMonitorConfig, i);
+        SpiceHead *head = &config->heads[i];
+        SPICE_DEBUG("monitor id: %u, surface id: %u, +%u+%u-%ux%u",
+                    head->id, head->surface_id,
+                    head->x, head->y, head->width, head->height);
+        mc->id = head->id;
+        mc->surface_id = head->surface_id;
+        mc->x = head->x;
+        mc->y = head->y;
+        mc->width = head->width;
+        mc->height = head->height;
+    }
+
+    g_object_notify_main_context(G_OBJECT(channel), "monitors");
+}
+
 static const spice_msg_handler display_handlers[] = {
     [ SPICE_MSG_DISPLAY_MODE ]               = display_handle_mode,
     [ SPICE_MSG_DISPLAY_MARK ]               = display_handle_mark,
@@ -1324,6 +1540,7 @@ static const spice_msg_handler display_handlers[] = {
     [ SPICE_MSG_DISPLAY_STREAM_CLIP ]        = display_handle_stream_clip,
     [ SPICE_MSG_DISPLAY_STREAM_DESTROY ]     = display_handle_stream_destroy,
     [ SPICE_MSG_DISPLAY_STREAM_DESTROY_ALL ] = display_handle_stream_destroy_all,
+    [ SPICE_MSG_DISPLAY_STREAM_DATA_SIZED ]  = display_handle_stream_data,
 
     [ SPICE_MSG_DISPLAY_DRAW_FILL ]          = display_handle_draw_fill,
     [ SPICE_MSG_DISPLAY_DRAW_OPAQUE ]        = display_handle_draw_opaque,
@@ -1340,6 +1557,8 @@ static const spice_msg_handler display_handlers[] = {
 
     [ SPICE_MSG_DISPLAY_SURFACE_CREATE ]     = display_handle_surface_create,
     [ SPICE_MSG_DISPLAY_SURFACE_DESTROY ]    = display_handle_surface_destroy,
+
+    [ SPICE_MSG_DISPLAY_MONITORS_CONFIG ]    = display_handle_monitors_config,
 };
 
 /* coroutine context */
