@@ -21,6 +21,19 @@
 
 #include "gio-coroutine.h"
 
+typedef struct _GConditionWaitSource
+{
+    GCoroutine *self;
+    GSource src;
+    GConditionWaitFunc func;
+    gpointer data;
+} GConditionWaitSource;
+
+GCoroutine* g_coroutine_self(void)
+{
+    return (GCoroutine*)coroutine_self();
+}
+
 /* Main loop helper functions */
 static gboolean g_io_wait_helper(GSocket *sock G_GNUC_UNUSED,
 				 GIOCondition cond,
@@ -31,55 +44,51 @@ static gboolean g_io_wait_helper(GSocket *sock G_GNUC_UNUSED,
     return FALSE;
 }
 
-GIOCondition g_io_wait(GSocket *sock, GIOCondition cond)
-{
-    GIOCondition *ret;
-    GSource *src = g_socket_create_source(sock,
-                                          cond | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
-                                          NULL);
-    g_source_set_callback(src, (GSourceFunc)g_io_wait_helper, coroutine_self(), NULL);
-    g_source_attach(src, NULL);
-    ret = coroutine_yield(NULL);
-    g_source_unref(src);
-    return *ret;
-}
-
-
-GIOCondition g_io_wait_interruptible(struct wait_queue *wait,
+GIOCondition g_coroutine_socket_wait(GCoroutine *self,
                                      GSocket *sock,
                                      GIOCondition cond)
 {
-    GIOCondition *ret;
-    gint id;
+    GIOCondition *ret, val = 0;
+    GSource *src;
 
-    g_return_val_if_fail(wait != NULL, 0);
+    g_return_val_if_fail(self != NULL, 0);
+    g_return_val_if_fail(self->wait_id == 0, 0);
     g_return_val_if_fail(sock != NULL, 0);
 
-    wait->context = coroutine_self();
-    GSource *src = g_socket_create_source(sock,
-                                          cond | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
-                                          NULL);
-    g_source_set_callback(src, (GSourceFunc)g_io_wait_helper,
-                          wait->context, NULL);
-    id = g_source_attach(src, NULL);
-    wait->waiting = TRUE;
+    src = g_socket_create_source(sock, cond | G_IO_HUP | G_IO_ERR | G_IO_NVAL, NULL);
+    g_source_set_callback(src, (GSourceFunc)g_io_wait_helper, self, NULL);
+    self->wait_id = g_source_attach(src, NULL);
     ret = coroutine_yield(NULL);
-    wait->waiting = FALSE;
     g_source_unref(src);
 
-    if (ret == NULL) {
-        g_source_remove(id);
-        return 0;
-    } else
-        return *ret;
+    if (ret != NULL)
+        val = *ret;
+    else
+        g_source_remove(self->wait_id);
+
+    self->wait_id = 0;
+    return val;
 }
 
-void g_io_wakeup(struct wait_queue *wait)
+void g_coroutine_condition_cancel(GCoroutine *coroutine)
 {
-    if (wait->waiting)
-        coroutine_yieldto(wait->context, NULL);
+    g_return_if_fail(coroutine != NULL);
+
+    if (coroutine->condition_id == 0)
+        return;
+
+    g_source_remove(coroutine->condition_id);
+    coroutine->condition_id = 0;
 }
 
+void g_coroutine_wakeup(GCoroutine *coroutine)
+{
+    g_return_if_fail(coroutine != NULL);
+    g_return_if_fail(coroutine != g_coroutine_self());
+
+    if (coroutine->wait_id)
+        coroutine_yieldto(&coroutine->coroutine, NULL);
+}
 
 /*
  * Call immediately before the main loop does an iteration. Returns
@@ -87,7 +96,7 @@ void g_io_wakeup(struct wait_queue *wait)
  */
 static gboolean g_condition_wait_prepare(GSource *src,
 					 int *timeout) {
-    struct g_condition_wait_source *vsrc = (struct g_condition_wait_source *)src;
+    GConditionWaitSource *vsrc = (GConditionWaitSource *)src;
     *timeout = -1;
     return vsrc->func(vsrc->data);
 }
@@ -98,7 +107,7 @@ static gboolean g_condition_wait_prepare(GSource *src,
  */
 static gboolean g_condition_wait_check(GSource *src)
 {
-    struct g_condition_wait_source *vsrc = (struct g_condition_wait_source *)src;
+    GConditionWaitSource *vsrc = (GConditionWaitSource *)src;
     return vsrc->func(vsrc->data);
 }
 
@@ -116,37 +125,59 @@ GSourceFuncs waitFuncs = {
 
 static gboolean g_condition_wait_helper(gpointer data)
 {
-    struct coroutine *co = (struct coroutine *)data;
-    coroutine_yieldto(co, NULL);
+    GCoroutine *self = (GCoroutine *)data;
+    coroutine_yieldto(&self->coroutine, NULL);
     return FALSE;
 }
 
-gboolean g_condition_wait(g_condition_wait_func func, gpointer data)
+/*
+ * g_coroutine_condition_wait:
+ * @coroutine: the coroutine to wait on
+ * @func: the condition callback
+ * @data: the user data passed to @func callback
+ *
+ * This function will wait on caller coroutine until @func returns %TRUE.
+ *
+ * @func is called when entering the main loop from the main context (coroutine).
+ *
+ * The condition can be cancelled by calling g_coroutine_wakeup()
+ *
+ * Returns: %TRUE if condition reached, %FALSE if not and cancelled
+ */
+gboolean g_coroutine_condition_wait(GCoroutine *self, GConditionWaitFunc func, gpointer data)
 {
     GSource *src;
-    struct g_condition_wait_source *vsrc;
+    GConditionWaitSource *vsrc;
+
+    g_return_val_if_fail(self != NULL, FALSE);
+    g_return_val_if_fail(self->condition_id == 0, FALSE);
+    g_return_val_if_fail(func != NULL, FALSE);
 
     /* Short-circuit check in case we've got it ahead of time */
-    if (func(data)) {
+    if (func(data))
         return TRUE;
-    }
 
     /*
      * Don't have it, so yield to the main loop, checking the condition
      * on each iteration of the main loop
      */
-    src = g_source_new(&waitFuncs, sizeof(struct g_condition_wait_source));
-    vsrc = (struct g_condition_wait_source *)src;
+    src = g_source_new(&waitFuncs, sizeof(GConditionWaitSource));
+    vsrc = (GConditionWaitSource *)src;
 
     vsrc->func = func;
     vsrc->data = data;
-    vsrc->co = coroutine_self();
+    vsrc->self = self;
 
-    g_source_attach(src, NULL);
-    g_source_set_callback(src, g_condition_wait_helper, coroutine_self(), NULL);
+    self->condition_id = g_source_attach(src, NULL);
+    g_source_set_callback(src, g_condition_wait_helper, self, NULL);
     coroutine_yield(NULL);
     g_source_unref(src);
 
+    /* it got woked up / cancelled? */
+    if (self->condition_id == 0)
+        return func(data);
+
+    self->condition_id = 0;
     return TRUE;
 }
 
@@ -158,6 +189,7 @@ struct signal_data
     gpointer params;
     GSignalEmitMainFunc func;
     const char *debug_info;
+    gboolean notified;
 };
 
 static gboolean emit_main_context(gpointer opaque)
@@ -165,6 +197,8 @@ static gboolean emit_main_context(gpointer opaque)
     struct signal_data *signal = opaque;
 
     signal->func(signal->object, signal->signum, signal->params);
+    signal->notified = TRUE;
+
     coroutine_yieldto(signal->caller, NULL);
 
     return FALSE;
@@ -179,12 +213,15 @@ void g_signal_emit_main_context(GObject *object,
 {
     struct signal_data data;
 
+    g_return_if_fail(coroutine_self()->caller);
+
     data.object = object;
     data.caller = coroutine_self();
     data.signum = signum;
     data.params = params;
     data.func = emit_main_func;
     data.debug_info = debug_info;
+    data.notified = FALSE;
     g_idle_add(emit_main_context, &data);
 
     /* This switches to the system coroutine context, lets
@@ -194,6 +231,7 @@ void g_signal_emit_main_context(GObject *object,
      * an idle function involved
      */
     coroutine_yield(NULL);
+    g_warn_if_fail(data.notified);
 }
 
 static gboolean notify_main_context(gpointer opaque)
@@ -201,6 +239,8 @@ static gboolean notify_main_context(gpointer opaque)
     struct signal_data *signal = opaque;
 
     g_object_notify(signal->object, signal->params);
+    signal->notified = TRUE;
+
     coroutine_yieldto(signal->caller, NULL);
 
     return FALSE;
@@ -212,9 +252,12 @@ void g_object_notify_main_context(GObject *object,
 {
     struct signal_data data;
 
+    g_return_if_fail(coroutine_self()->caller);
+
     data.object = object;
     data.caller = coroutine_self();
     data.params = (gpointer)property_name;
+    data.notified = FALSE;
 
     g_idle_add(notify_main_context, &data);
 
@@ -225,4 +268,5 @@ void g_object_notify_main_context(GObject *object,
      * an idle function involved
      */
     coroutine_yield(NULL);
+    g_warn_if_fail(data.notified);
 }
