@@ -48,6 +48,7 @@ static void spice_channel_write_msg(SpiceChannel *channel, SpiceMsgOut *out);
 static void spice_channel_send_link(SpiceChannel *channel);
 static void channel_disconnect(SpiceChannel *channel);
 static void channel_reset(SpiceChannel *channel, gboolean migrating);
+static void spice_channel_reset_capabilities(SpiceChannel *channel);
 
 /**
  * SECTION:spice-channel
@@ -1147,7 +1148,11 @@ static void spice_channel_send_link(SpiceChannel *channel)
         *(uint32_t *)p = g_array_index(c->caps, uint32_t, i);
         p += sizeof(uint32_t);
     }
-
+    SPICE_DEBUG("channel type %d id %d num common caps %d num caps %d",
+                c->link_msg.channel_type,
+                c->link_msg.channel_id,
+                c->link_msg.num_common_caps,
+                c->link_msg.num_channel_caps);
     spice_channel_write(channel, buffer, p - buffer);
     free(buffer);
 }
@@ -2095,12 +2100,14 @@ static void *spice_channel_coroutine(void *data)
     if (spice_session_get_client_provided_socket(c->session)) {
         if (c->fd < 0) {
             g_critical("fd not provided!");
+            emit_main_context(channel, SPICE_CHANNEL_EVENT, SPICE_CHANNEL_ERROR_CONNECT);
             goto cleanup;
         }
 
         if (!(c->sock = g_socket_new_from_fd(c->fd, NULL))) {
                 SPICE_DEBUG("Failed to open socket from fd %d", c->fd);
-                return FALSE;
+                emit_main_context(channel, SPICE_CHANNEL_EVENT, SPICE_CHANNEL_ERROR_CONNECT);
+                goto cleanup;
         }
 
         g_socket_set_blocking(c->sock, FALSE);
@@ -2128,6 +2135,7 @@ reconnect:
         c->ctx = SSL_CTX_new(TLSv1_method());
         if (c->ctx == NULL) {
             g_critical("SSL_CTX_new failed");
+            emit_main_context(channel, SPICE_CHANNEL_EVENT, SPICE_CHANNEL_ERROR_TLS);
             goto cleanup;
         }
 
@@ -2139,15 +2147,16 @@ reconnect:
             g_warn_if_fail(ca_file != NULL);
             SPICE_DEBUG("CA file: %s", ca_file);
             rc = SSL_CTX_load_verify_locations(c->ctx, ca_file, NULL);
-            if (rc != 1)
-                g_warning("loading ca certs from %s failed", ca_file);
 
             if (rc != 1) {
+                g_warning("loading ca certs from %s failed", ca_file);
                 if (verify & SPICE_SESSION_VERIFY_PUBKEY) {
                     g_warning("only pubkey active");
                     verify = SPICE_SESSION_VERIFY_PUBKEY;
-                } else
+                } else {
+                    emit_main_context(channel, SPICE_CHANNEL_EVENT, SPICE_CHANNEL_ERROR_TLS);
                     goto cleanup;
+                }
             }
         }
 
@@ -2163,6 +2172,7 @@ reconnect:
         c->ssl = SSL_new(c->ctx);
         if (c->ssl == NULL) {
             g_critical("SSL_new failed");
+            emit_main_context(channel, SPICE_CHANNEL_EVENT, SPICE_CHANNEL_ERROR_TLS);
             goto cleanup;
         }
 
@@ -2304,7 +2314,7 @@ gboolean spice_channel_connect(SpiceChannel *channel)
     g_return_val_if_fail(SPICE_IS_CHANNEL(channel), FALSE);
     SpiceChannelPrivate *c = channel->priv;
 
-    if (c->state == SPICE_CHANNEL_STATE_CONNECTING)
+    if (c->state >= SPICE_CHANNEL_STATE_CONNECTING)
         return TRUE;
 
     return channel_connect(channel);
@@ -2390,10 +2400,10 @@ static void channel_reset(SpiceChannel *channel, gboolean migrating)
     g_array_set_size(c->remote_common_caps, 0);
     g_array_set_size(c->remote_caps, 0);
     g_array_set_size(c->common_caps, 0);
-    g_array_set_size(c->caps, 0);
     /* Restore our default capabilities in case the channel gets re-used */
     spice_channel_set_common_capability(channel, SPICE_COMMON_CAP_PROTOCOL_AUTH_SELECTION);
     spice_channel_set_common_capability(channel, SPICE_COMMON_CAP_MINI_HEADER);
+    spice_channel_reset_capabilities(channel);
 }
 
 /* system or coroutine context */
@@ -2532,8 +2542,9 @@ static void set_capability(GArray *caps, guint32 cap)
  * @cap: a capability
  *
  * Enable specific channel-kind capability.
+ * Deprecated: 0.13: this function has been removed
  **/
-/* FIXME: we may want to make caps read only from outside */
+#undef spice_channel_set_capability
 void spice_channel_set_capability(SpiceChannel *channel, guint32 cap)
 {
     SpiceChannelPrivate *c;
@@ -2545,14 +2556,15 @@ void spice_channel_set_capability(SpiceChannel *channel, guint32 cap)
 }
 
 G_GNUC_INTERNAL
-void spice_channel_set_common_capability(SpiceChannel *channel, guint32 cap)
+void spice_caps_set(GArray *caps, guint32 cap, const gchar *desc)
 {
-    SpiceChannelPrivate *c;
+    g_return_if_fail(caps != NULL);
+    g_return_if_fail(desc != NULL);
 
-    g_return_if_fail(SPICE_IS_CHANNEL(channel));
+    if (g_strcmp0(g_getenv(desc), "0") == 0)
+        return;
 
-    c = channel->priv;
-    set_capability(c->common_caps, cap);
+    set_capability(caps, cap);
 }
 
 G_GNUC_INTERNAL
@@ -2632,3 +2644,13 @@ static void spice_channel_handle_msg(SpiceChannel *channel, SpiceMsgIn *msg)
 
     base_handlers[type](channel, msg);
 }
+
+static void spice_channel_reset_capabilities(SpiceChannel *channel)
+{
+    SpiceChannelPrivate *c = SPICE_CHANNEL_GET_PRIVATE(channel);
+    g_array_set_size(c->caps, 0);
+
+    if (SPICE_CHANNEL_GET_CLASS(channel)->channel_reset_capabilities) {
+        SPICE_CHANNEL_GET_CLASS(channel)->channel_reset_capabilities(channel);
+    }
+}
