@@ -155,6 +155,7 @@ spice_session_dispose(GObject *gobject)
     }
 
     g_clear_object(&s->audio_manager);
+    g_clear_object(&s->desktop_integration);
     g_clear_object(&s->gtk_session);
     g_clear_object(&s->usb_manager);
 
@@ -249,7 +250,6 @@ static int spice_uri_parse(SpiceSession *session, const char *original_uri)
     gchar key[32], value[128];
     gchar *host = NULL, *port = NULL, *tls_port = NULL, *uri = NULL, *password = NULL;
     gchar **target_key;
-    int len;
     gchar *path = NULL;
     gchar *authority = NULL;
     gchar *query = NULL;
@@ -277,12 +277,10 @@ static int spice_uri_parse(SpiceSession *session, const char *original_uri)
 
     if (path) {
         size_t prefix = strcspn(path, URI_QUERY_START);
-        if (len)
-            query = path + prefix;
+        query = path + prefix;
     } else {
         size_t prefix = strcspn(authority, URI_QUERY_START);
-        if (len)
-            query = authority + prefix;
+        query = authority + prefix;
     }
 
     if (query && query[0]) {
@@ -320,6 +318,7 @@ static int spice_uri_parse(SpiceSession *session, const char *original_uri)
     }
 
     while (query && query[0] != '\0') {
+        int len;
         if (sscanf(query, "%31[-a-zA-Z0-9]=%127[^;&]%n", key, value, &len) != 2) {
             g_warning("Failed to parse URI query '%s'", query);
             goto fail;
@@ -521,13 +520,17 @@ static void spice_session_set_property(GObject      *gobject,
             g_byte_array_unref(s->pubkey);
         s->pubkey = g_value_dup_boxed(value);
         if (s->pubkey)
-            s->verify = SPICE_SESSION_VERIFY_PUBKEY;
+            s->verify |= SPICE_SESSION_VERIFY_PUBKEY;
+        else
+            s->verify &= ~SPICE_SESSION_VERIFY_PUBKEY;
 	break;
     case PROP_CERT_SUBJECT:
         g_free(s->cert_subject);
         s->cert_subject = g_value_dup_string(value);
         if (s->cert_subject)
-            s->verify = SPICE_SESSION_VERIFY_SUBJECT;
+            s->verify |= SPICE_SESSION_VERIFY_SUBJECT;
+        else
+            s->verify &= ~SPICE_SESSION_VERIFY_SUBJECT;
         break;
     case PROP_VERIFY:
         s->verify = g_value_get_flags(value);
@@ -1611,6 +1614,9 @@ void spice_session_channel_new(SpiceSession *session, SpiceChannel *channel)
                      NULL);
         if (s->color_depth != 0)
             g_object_set(channel, "color-depth", s->color_depth, NULL);
+
+        SPICE_DEBUG("new main channel, switching");
+        s->cmain = channel;
     }
 
     g_signal_emit(session, signals[SPICE_SESSION_CHANNEL_NEW], 0, channel);
@@ -1621,7 +1627,7 @@ void spice_session_channel_destroy(SpiceSession *session, SpiceChannel *channel)
 {
     SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
     struct channel *item = NULL;
-    RingItem *ring, *next;
+    RingItem *ring;
 
     g_return_if_fail(s != NULL);
     g_return_if_fail(channel != NULL);
@@ -1630,22 +1636,23 @@ void spice_session_channel_destroy(SpiceSession *session, SpiceChannel *channel)
         s->migration_left = g_list_remove(s->migration_left, channel);
 
     for (ring = ring_get_head(&s->channels); ring != NULL;
-         ring = next) {
-        next = ring_next(&s->channels, ring);
+         ring = ring_next(&s->channels, ring)) {
         item = SPICE_CONTAINEROF(ring, struct channel, link);
-        if (item->channel == s->cmain) {
-            SPICE_DEBUG("the session lost the main channel");
-            s->cmain = NULL;
-        }
-        if (item->channel == channel) {
-            ring_remove(&item->link);
-            free(item);
-            g_signal_emit(session, signals[SPICE_SESSION_CHANNEL_DESTROY], 0, channel);
-            return;
-        }
+        if (item->channel == channel)
+            break;
     }
 
-    g_warn_if_reached();
+    g_return_if_fail(ring != NULL);
+
+    if (channel == s->cmain) {
+        SPICE_DEBUG("the session lost the main channel");
+        s->cmain = NULL;
+    }
+
+    ring_remove(&item->link);
+    free(item);
+
+    g_signal_emit(session, signals[SPICE_SESSION_CHANNEL_DESTROY], 0, channel);
 }
 
 G_GNUC_INTERNAL
