@@ -101,7 +101,9 @@ enum {
     PROP_AUTO_CLIPBOARD,
     PROP_SCALING,
     PROP_DISABLE_INPUTS,
-    PROP_ZOOM_LEVEL
+    PROP_ZOOM_LEVEL,
+    PROP_MONITOR_ID,
+    PROP_READY
 };
 
 /* Signals */
@@ -125,13 +127,11 @@ static void update_mouse_grab(SpiceDisplay *display);
 static void try_mouse_grab(SpiceDisplay *display);
 static void try_mouse_ungrab(SpiceDisplay *display);
 static void recalc_geometry(GtkWidget *widget);
-static void disconnect_main(SpiceDisplay *display);
-static void disconnect_cursor(SpiceDisplay *display);
-static void disconnect_display(SpiceDisplay *display);
 static void channel_new(SpiceSession *s, SpiceChannel *channel, gpointer data);
 static void channel_destroy(SpiceSession *s, SpiceChannel *channel, gpointer data);
 static void sync_keyboard_lock_modifiers(SpiceDisplay *display);
 static void cursor_invalidate(SpiceDisplay *display);
+static void update_area(SpiceDisplay *display, gint x, gint y, gint width, gint height);
 
 /* ---------------------------------------------------------------- */
 
@@ -151,6 +151,9 @@ static void spice_display_get_property(GObject    *object,
     case PROP_CHANNEL_ID:
         g_value_set_int(value, d->channel_id);
         break;
+    case PROP_MONITOR_ID:
+        g_value_set_int(value, d->monitor_id);
+        break;
     case PROP_KEYBOARD_GRAB:
         g_value_set_boolean(value, d->keyboard_grab_enable);
         break;
@@ -173,6 +176,9 @@ static void spice_display_get_property(GObject    *object,
     case PROP_ZOOM_LEVEL:
         g_value_set_int(value, d->zoom_level);
         break;
+    case PROP_READY:
+        g_value_set_boolean(value, d->ready);
+        break;
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
         break;
@@ -201,14 +207,94 @@ static void update_size_request(SpiceDisplay *display)
         reqwidth = 640;
         reqheight = 480;
     } else {
-        reqwidth = d->width;
-        reqheight = d->height;
+        reqwidth = d->area.width;
+        reqheight = d->area.height;
     }
 
     gtk_widget_set_size_request(GTK_WIDGET(display), reqwidth, reqheight);
     recalc_geometry(GTK_WIDGET(display));
 }
 
+static void update_keyboard_focus(SpiceDisplay *display, gboolean state)
+{
+    SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
+
+    d->keyboard_have_focus = state;
+
+    /* keyboard grab gets inhibited by usb-device-manager when it is
+       in the process of redirecting a usb-device (as this may show a
+       policykit dialog). Making autoredir/automount setting changes while
+       this is happening is not a good idea! */
+    if (d->keyboard_grab_inhibit)
+        return;
+
+    spice_gtk_session_request_auto_usbredir(d->gtk_session, state);
+}
+
+static void update_ready(SpiceDisplay *display)
+{
+    SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
+    gboolean ready;
+
+    ready = d->mark != 0 && d->monitor_ready;
+
+    if (d->ready == ready)
+        return;
+
+    if (ready && gtk_widget_get_window(GTK_WIDGET(display)))
+        gtk_widget_queue_draw(GTK_WIDGET(display));
+
+    d->ready = ready;
+    g_object_notify(G_OBJECT(display), "ready");
+}
+
+static void set_monitor_ready(SpiceDisplay *self, gboolean ready)
+{
+    SpiceDisplayPrivate *d = self->priv;
+
+    d->monitor_ready = ready;
+    update_ready(self);
+}
+
+static void update_monitor_area(SpiceDisplay *display)
+{
+    SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
+    SpiceDisplayMonitorConfig *c;
+    GArray *monitors = NULL;
+
+    SPICE_DEBUG("update monitor area %d:%d", d->channel_id, d->monitor_id);
+    if (d->monitor_id < 0)
+        goto whole;
+
+    g_object_get(d->display, "monitors", &monitors, NULL);
+    if (monitors == NULL || d->monitor_id >= monitors->len) {
+        SPICE_DEBUG("update monitor: no monitor %d", d->monitor_id);
+        set_monitor_ready(display, false);
+        if (spice_channel_test_capability(d->display, SPICE_DISPLAY_CAP_MONITORS_CONFIG)) {
+            SPICE_DEBUG("waiting until MonitorsConfig is received");
+            return;
+        }
+        goto whole;
+    }
+
+    c = &g_array_index(monitors, SpiceDisplayMonitorConfig, d->monitor_id);
+    g_return_if_fail(c != NULL);
+    if (c->surface_id != 0) {
+        g_warning("FIXME: only support monitor config with primary surface 0, "
+                  "but given config surface %d", c->surface_id);
+        goto whole;
+    }
+
+    update_area(display, c->x, c->y, c->width, c->height);
+    g_clear_pointer(&monitors, g_array_unref);
+    return;
+
+whole:
+    /* by display whole surface */
+    update_area(display, 0, 0, d->width, d->height);
+    set_monitor_ready(display, true);
+}
+
 static void spice_display_set_property(GObject      *object,
                                        guint         prop_id,
                                        const GValue *value,
@@ -219,12 +305,18 @@ static void spice_display_set_property(GObject      *object,
 
     switch (prop_id) {
     case PROP_SESSION:
-        d->session = g_object_ref(g_value_get_object(value));
+        g_warn_if_fail(d->session == NULL);
+        d->session = g_value_dup_object(value);
         d->gtk_session = spice_gtk_session_get(d->session);
         break;
     case PROP_CHANNEL_ID:
         d->channel_id = g_value_get_int(value);
         break;
+    case PROP_MONITOR_ID:
+        d->monitor_id = g_value_get_int(value);
+        if (d->display) /* if constructed */
+            update_monitor_area(display);
+        break;
     case PROP_KEYBOARD_GRAB:
         d->keyboard_grab_enable = g_value_get_boolean(value);
         update_keyboard_grab(display);
@@ -289,19 +381,9 @@ static void spice_display_dispose(GObject *obj)
 
     SPICE_DEBUG("spice display dispose");
 
-    disconnect_main(display);
-    disconnect_display(display);
-    disconnect_cursor(display);
-
-    if (d->session) {
-        g_signal_handlers_disconnect_by_func(d->session, G_CALLBACK(channel_new),
-                                             display);
-        g_signal_handlers_disconnect_by_func(d->session, G_CALLBACK(channel_destroy),
-                                             display);
-        g_object_unref(d->session);
-        d->session = NULL;
-        d->gtk_session = NULL;
-    }
+    spicex_image_destroy(display);
+    g_clear_object(&d->session);
+    d->gtk_session = NULL;
 
     G_OBJECT_CLASS(spice_display_parent_class)->dispose(obj);
 }
@@ -416,30 +498,37 @@ spice_display_constructor(GType                  gtype,
     if (!d->session)
         g_error("SpiceDisplay constructed without a session");
 
-    g_signal_connect(d->session, "channel-new",
-                     G_CALLBACK(channel_new), display);
-    g_signal_connect(d->session, "channel-destroy",
-                     G_CALLBACK(channel_destroy), display);
+    spice_g_signal_connect_object(d->session, "channel-new",
+                                  G_CALLBACK(channel_new), display, 0);
+    spice_g_signal_connect_object(d->session, "channel-destroy",
+                                  G_CALLBACK(channel_destroy), display, 0);
     list = spice_session_get_channels(d->session);
     for (it = g_list_first(list); it != NULL; it = g_list_next(it)) {
-        channel_new(d->session, it->data, (gpointer*)display);
+        if (SPICE_IS_MAIN_CHANNEL(it->data)) {
+            channel_new(d->session, it->data, (gpointer*)display);
+            break;
+        }
+    }
+    for (it = g_list_first(list); it != NULL; it = g_list_next(it)) {
+        if (!SPICE_IS_MAIN_CHANNEL(it->data))
+            channel_new(d->session, it->data, (gpointer*)display);
     }
     g_list_free(list);
 
     spice_g_signal_connect_object(d->gtk_session, "notify::auto-clipboard",
                                   G_CALLBACK(gtk_session_property_changed), display, 0);
 
-    g_signal_connect(d->session, "notify::inhibit-keyboard-grab",
-                     G_CALLBACK(session_inhibit_keyboard_grab_changed),
-                     display);
+    spice_g_signal_connect_object(d->session, "notify::inhibit-keyboard-grab",
+                                  G_CALLBACK(session_inhibit_keyboard_grab_changed),
+                                  display, 0);
 
     return obj;
 }
 
 /**
  * spice_display_set_grab_keys:
- * @display:
- * @seq: key sequence
+ * @display: the display widget
+ * @seq: (transfer none): key sequence
  *
  * Set the key combination to grab/ungrab the keyboard. The default is
  * "Control L + Alt L".
@@ -493,9 +582,9 @@ static LRESULT CALLBACK keyboard_hook_cb(int code, WPARAM wparam, LPARAM lparam)
 
 /**
  * spice_display_get_grab_keys:
- * @display:
+ * @display: the display widget
  *
- * Returns: the current grab key combination.
+ * Returns: (transfer none): the current grab key combination.
  **/
 SpiceGrabSequence *spice_display_get_grab_keys(SpiceDisplay *display)
 {
@@ -761,6 +850,19 @@ static void update_mouse_grab(SpiceDisplay *display)
         try_mouse_ungrab(display);
 }
 
+static gint get_display_id(SpiceDisplay *display)
+{
+    SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
+
+    /* supported monitor_id only with display channel #0 */
+    if (d->channel_id == 0 && d->monitor_id >= 0)
+        return d->monitor_id;
+
+    g_return_val_if_fail(d->monitor_id <= 0, -1);
+
+    return d->channel_id;
+}
+
 static void recalc_geometry(GtkWidget *widget)
 {
     SpiceDisplay *display = SPICE_DISPLAY(widget);
@@ -770,19 +872,20 @@ static void recalc_geometry(GtkWidget *widget)
     d->mx = 0;
     d->my = 0;
     if (!spicex_is_scaled(display)) {
-        if (d->ww > d->width)
-            d->mx = (d->ww - d->width) / 2;
-        if (d->wh > d->height)
-            d->my = (d->wh - d->height) / 2;
+        if (d->ww > d->area.width)
+            d->mx = (d->ww - d->area.width) / 2;
+        if (d->wh > d->area.height)
+            d->my = (d->wh - d->area.height) / 2;
     } else
         zoom = (gdouble)d->zoom_level / 100;
 
-    SPICE_DEBUG("monitors: id %d, guest %dx%d, window %dx%d, zoom %g, offset +%d+%d",
-                d->channel_id, d->width, d->height, d->ww, d->wh, zoom, d->mx, d->my);
+    SPICE_DEBUG("recalc geom monitor: %d:%d, guest +%d+%d:%dx%d, window %dx%d, zoom %g, offset +%d+%d",
+                d->channel_id, d->monitor_id, d->area.x, d->area.y, d->area.width, d->area.height,
+                d->ww, d->wh, zoom, d->mx, d->my);
 
     if (d->resize_guest_enable)
-        spice_main_set_display(d->main, d->channel_id,
-                               0, 0, d->ww / zoom, d->wh / zoom);
+        spice_main_set_display(d->main, get_display_id(display),
+                               d->area.x, d->area.y, d->ww / zoom, d->wh / zoom);
 }
 
 /* ---------------------------------------------------------------- */
@@ -801,44 +904,36 @@ static void recalc_geometry(GtkWidget *widget)
 
 #define CONVERT_0555_TO_8888(s) (CONVERT_0555_TO_0888(s) | 0xff000000)
 
-static gboolean do_color_convert(SpiceDisplay *display,
-                                 gint x, gint y, gint w, gint h)
+static gboolean do_color_convert(SpiceDisplay *display, GdkRectangle *r)
 {
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
-    int i, j, maxy, maxx, miny, minx;
     guint32 *dest = d->data;
     guint16 *src = d->data_origin;
+    gint x, y;
 
-    if (!d->convert)
-        return true;
-
+    g_return_val_if_fail(r != NULL, false);
     g_return_val_if_fail(d->format == SPICE_SURFACE_FMT_16_555 ||
                          d->format == SPICE_SURFACE_FMT_16_565, false);
 
-    miny = MAX(y, 0);
-    minx = MAX(x, 0);
-    maxy = MIN(y + h, d->height);
-    maxx = MIN(x + w, d->width);
-
-    dest +=  (d->stride / 4) * miny;
-    src += (d->stride / 2) * miny;
+    src += (d->stride / 2) * r->y + r->x;
+    dest += d->area.width * (r->y - d->area.y) + (r->x - d->area.x);
 
     if (d->format == SPICE_SURFACE_FMT_16_555) {
-        for (j = miny; j < maxy; j++) {
-            for (i = minx; i < maxx; i++) {
-                dest[i] = CONVERT_0555_TO_0888(src[i]);
+        for (y = 0; y < r->height; y++) {
+            for (x = 0; x < r->width; x++) {
+                dest[x] = CONVERT_0555_TO_0888(src[x]);
             }
 
-            dest += d->stride / 4;
+            dest += d->area.width;
             src += d->stride / 2;
         }
     } else if (d->format == SPICE_SURFACE_FMT_16_565) {
-        for (j = miny; j < maxy; j++) {
-            for (i = minx; i < maxx; i++) {
-                dest[i] = CONVERT_0565_TO_0888(src[i]);
+        for (y = 0; y < r->height; y++) {
+            for (x = 0; x < r->width; x++) {
+                dest[x] = CONVERT_0565_TO_0888(src[x]);
             }
 
-            dest += d->stride / 4;
+            dest += d->area.width;
             src += d->stride / 2;
         }
     }
@@ -854,12 +949,10 @@ static gboolean draw_event(GtkWidget *widget, cairo_t *cr)
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
     g_return_val_if_fail(d != NULL, false);
 
-    if (d->mark == 0 || d->data == NULL)
+    if (d->mark == 0 || d->data == NULL ||
+        d->area.width == 0 || d->area.height == 0)
         return false;
-
-    if (!d->ximage) {
-        spicex_image_create(display);
-    }
+    g_return_val_if_fail(d->ximage != NULL, false);
 
     spicex_draw_event(display, cr);
     update_mouse_pointer(display);
@@ -873,12 +966,10 @@ static gboolean expose_event(GtkWidget *widget, GdkEventExpose *expose)
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
     g_return_val_if_fail(d != NULL, false);
 
-    if (d->mark == 0 || d->data == NULL)
+    if (d->mark == 0 || d->data == NULL ||
+        d->area.width == 0 || d->area.height == 0)
         return false;
-
-    if (!d->ximage) {
-        spicex_image_create(display);
-    }
+    g_return_val_if_fail(d->ximage != NULL, false);
 
     spicex_expose_event(display, expose);
     update_mouse_pointer(display);
@@ -941,11 +1032,7 @@ static gboolean check_for_grab_key(SpiceDisplay *display, int type, int keyval)
     if (!d->grabseq->nkeysyms)
         return FALSE;
 
-    if (type == GDK_KEY_RELEASE) {
-        /* Any key release resets the whole grab sequence */
-        memset(d->activeseq, 0, sizeof(gboolean) * d->grabseq->nkeysyms);
-        return FALSE;
-    } else {
+    if (type == GDK_KEY_PRESS) {
         /* Record the new key press */
         for (i = 0 ; i < d->grabseq->nkeysyms ; i++)
             if (d->grabseq->keysyms[i] == keyval)
@@ -956,8 +1043,17 @@ static gboolean check_for_grab_key(SpiceDisplay *display, int type, int keyval)
             if (d->activeseq[i] == FALSE)
                 return FALSE;
 
+        /* resets the whole grab sequence on success */
+        memset(d->activeseq, 0, sizeof(gboolean) * d->grabseq->nkeysyms);
         return TRUE;
-    }
+    } else if (type == GDK_KEY_RELEASE) {
+        /* Any key release resets the whole grab sequence */
+        memset(d->activeseq, 0, sizeof(gboolean) * d->grabseq->nkeysyms);
+        return FALSE;
+    } else
+        g_warn_if_reached();
+
+    return FALSE;
 }
 
 static gboolean key_event(GtkWidget *widget, GdkEventKey *key)
@@ -980,7 +1076,10 @@ static gboolean key_event(GtkWidget *widget, GdkEventKey *key)
                 try_mouse_grab(display);
         }
 
-        return true;
+        // that's the last key pressed from the grab sequence
+        // let send it to the remote if it's a modifier key
+        if (!key->is_modifier)
+            return true;
     }
 
     if (!d->inputs)
@@ -996,6 +1095,7 @@ static gboolean key_event(GtkWidget *widget, GdkEventKey *key)
         send_key(display, scancode, 0);
         break;
     default:
+        g_warn_if_reached();
         break;
     }
 
@@ -1083,10 +1183,8 @@ static gboolean focus_in_event(GtkWidget *widget, GdkEventFocus *focus G_GNUC_UN
 
     release_keys(display);
     sync_keyboard_lock_modifiers(display);
-    d->keyboard_have_focus = true;
+    update_keyboard_focus(display, true);
     try_keyboard_grab(display);
-    spice_gtk_session_update_keyboard_focus(d->gtk_session,
-                                            d->keyboard_have_focus);
 #ifdef WIN32
     focus_window = GDK_WINDOW_HWND(gtk_widget_get_window(widget));
     g_return_val_if_fail(focus_window != NULL, true);
@@ -1109,9 +1207,8 @@ static gboolean focus_out_event(GtkWidget *widget, GdkEventFocus *focus G_GNUC_U
         return true;
 
     release_keys(display);
-    d->keyboard_have_focus = false;
-    spice_gtk_session_update_keyboard_focus(d->gtk_session,
-                                            d->keyboard_have_focus);
+    update_keyboard_focus(display, false);
+
     return true;
 }
 
@@ -1156,10 +1253,10 @@ static gboolean motion_event(GtkWidget *widget, GdkEventMotion *motion)
         return true;
 
     gdk_drawable_get_size(gtk_widget_get_window(widget), &ww, &wh);
-    if (spicex_is_scaled(display) && (d->width != ww || d->height != wh)) {
+    if (spicex_is_scaled(display) && (d->area.width != ww || d->area.height != wh)) {
         double sx, sy;
-        sx = (double)d->width / (double)ww;
-        sy = (double)d->height / (double)wh;
+        sx = (double)d->area.width / (double)ww;
+        sy = (double)d->area.height / (double)wh;
 
         /* Scaling the desktop, so scale the mouse coords by the same.
          * Ratio for the rounding used:
@@ -1183,10 +1280,10 @@ static gboolean motion_event(GtkWidget *widget, GdkEventMotion *motion)
 
     switch (d->mouse_mode) {
     case SPICE_MOUSE_MODE_CLIENT:
-        if (motion->x >= 0 && motion->x < d->width &&
-            motion->y >= 0 && motion->y < d->height) {
+        if (motion->x >= 0 && motion->x < d->area.width &&
+            motion->y >= 0 && motion->y < d->area.height) {
             spice_inputs_position(d->inputs,
-                                  motion->x, motion->y,
+                                  motion->x + d->area.x, motion->y + d->area.y,
                                   d->channel_id,
                                   button_mask_gdk_to_spice(motion->state));
         }
@@ -1260,7 +1357,7 @@ static gboolean button_event(GtkWidget *widget, GdkEventButton *button)
         gtk_widget_get_pointer (widget, &x, &y);
         x -= d->mx;
         y -= d->my;
-        if (!(x >= 0 && x < d->width && y >= 0 && y < d->height))
+        if (!(x >= 0 && x < d->area.width && y >= 0 && y < d->area.height))
             return true;
     }
 
@@ -1409,6 +1506,24 @@ static void spice_display_class_init(SpiceDisplayClass *klass)
                               G_PARAM_CONSTRUCT |
                               G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceDisplay:ready:
+     *
+     * Indicate whether the display is ready to be shown. It takes
+     * into account several conditions, such as the channel display
+     * "mark" state, whether the monitor area is visible..
+     *
+     * Since: 0.13
+     **/
+    g_object_class_install_property
+        (gobject_class, PROP_READY,
+         g_param_spec_boolean("ready",
+                              "Ready",
+                              "Ready to display",
+                              FALSE,
+                              G_PARAM_READABLE |
+                              G_PARAM_STATIC_STRINGS));
+
     /**
      * SpiceDisplay:auto-clipboard:
      *
@@ -1472,6 +1587,25 @@ static void spice_display_class_init(SpiceDisplayClass *klass)
                           G_PARAM_CONSTRUCT |
                           G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceDisplay:monitor-id:
+     *
+     * Select monitor from #SpiceDisplay to show.
+     * The value -1 means the whole display is shown.
+     * By default, the monitor 0 is selected.
+     *
+     * Since: 0.13
+     **/
+    g_object_class_install_property
+        (gobject_class, PROP_MONITOR_ID,
+         g_param_spec_int("monitor-id",
+                          "Monitor ID",
+                          "Select monitor ID",
+                          -1, G_MAXINT, 0,
+                          G_PARAM_READWRITE |
+                          G_PARAM_CONSTRUCT |
+                          G_PARAM_STATIC_STRINGS));
+
     /**
      * SpiceDisplay::mouse-grab:
      * @display: the #SpiceDisplay that emitted the signal
@@ -1552,6 +1686,40 @@ static void update_mouse_mode(SpiceChannel *channel, gpointer data)
     cursor_invalidate(display);
 }
 
+static void update_area(SpiceDisplay *display,
+                        gint x, gint y, gint width, gint height)
+{
+    SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
+    GdkRectangle primary = {
+        .x = 0,
+        .y = 0,
+        .width = d->width,
+        .height = d->height
+    };
+    GdkRectangle area = {
+        .x = x,
+        .y = y,
+        .width = width,
+        .height = height
+    };
+
+    SPICE_DEBUG("update area, primary: %dx%d, area: +%d+%d %dx%d", d->width, d->height, area.x, area.y, area.width, area.height);
+
+    if (!gdk_rectangle_intersect(&primary, &area, &area)) {
+        SPICE_DEBUG("The monitor area is not intersecting primary surface");
+        memset(&d->area, '\0', sizeof(d->area));
+        set_monitor_ready(display, false);
+        return;
+    }
+
+    spicex_image_destroy(display);
+    d->area = area;
+    spicex_image_create(display);
+    update_size_request(display);
+
+    set_monitor_ready(display, true);
+}
+
 static void primary_create(SpiceChannel *channel, gint format,
                            gint width, gint height, gint stride,
                            gint shmid, gpointer imgdata, gpointer data)
@@ -1561,14 +1729,12 @@ static void primary_create(SpiceChannel *channel, gint format,
 
     d->format = format;
     d->stride = stride;
-    d->shmid  = shmid;
+    d->shmid = shmid;
+    d->width = width;
+    d->height = height;
     d->data_origin = d->data = imgdata;
 
-    if (d->width != width || d->height != height) {
-        d->width  = width;
-        d->height = height;
-        update_size_request(display);
-    }
+    update_monitor_area(display);
 }
 
 static void primary_destroy(SpiceChannel *channel, gpointer data)
@@ -1577,43 +1743,54 @@ static void primary_destroy(SpiceChannel *channel, gpointer data)
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
 
     spicex_image_destroy(display);
-    d->format = 0;
     d->width  = 0;
     d->height = 0;
     d->stride = 0;
     d->shmid  = 0;
-    d->data   = 0;
-    d->data_origin = 0;
+    d->data = NULL;
+    d->data_origin = NULL;
+    set_monitor_ready(display, false);
 }
 
 static void invalidate(SpiceChannel *channel,
                        gint x, gint y, gint w, gint h, gpointer data)
 {
     SpiceDisplay *display = data;
+    SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
+    GdkRectangle rect = {
+        .x = x,
+        .y = y,
+        .width = w,
+        .height = h
+    };
 
     if (!gtk_widget_get_window(GTK_WIDGET(display)))
         return;
 
-    if (!do_color_convert(display, x, y, w, h))
+    if (!gdk_rectangle_intersect(&rect, &d->area, &rect))
         return;
 
+    if (d->convert)
+        do_color_convert(display, &rect);
+
+    x = rect.x - d->area.x;
+    y = rect.y - d->area.y;
+    w = rect.width;
+    h = rect.height;
+
     spicex_image_invalidate(display, &x, &y, &w, &h);
     gtk_widget_queue_draw_area(GTK_WIDGET(display),
                                x, y, w, h);
 }
 
-static void mark(SpiceChannel *channel, gint mark, gpointer data)
+static void mark(SpiceDisplay *display, gint mark)
 {
-    SpiceDisplay *display = data;
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
     g_return_if_fail(d != NULL);
 
-    SPICE_DEBUG("widget mark: %d, channel %d", mark, d->channel_id);
+    SPICE_DEBUG("widget mark: %d, %d:%d %p", mark, d->channel_id, d->monitor_id, display);
     d->mark = mark;
-    spice_main_set_display_enabled(d->main, d->channel_id, d->mark != 0);
-    if (mark != 0 && gtk_widget_get_window(GTK_WIDGET(display)))
-        gdk_window_invalidate_rect(gtk_widget_get_window(GTK_WIDGET(display)),
-                                   NULL, FALSE);
+    update_ready(display);
 }
 
 static void cursor_set(SpiceCursorChannel *channel,
@@ -1682,7 +1859,7 @@ G_GNUC_INTERNAL
 void spice_display_get_scaling(SpiceDisplay *display, double *sx, double *sy)
 {
     SpiceDisplayPrivate *d = display->priv;
-    int fbw = d->width, fbh = d->height;
+    int fbw = d->area.width, fbh = d->area.height;
     int ww, wh;
 
     if (!spicex_is_scaled(display) || !gtk_widget_get_window(GTK_WIDGET(display))) {
@@ -1708,8 +1885,8 @@ static void cursor_invalidate(SpiceDisplay *display)
     spice_display_get_scaling(display, &sx, &sy);
 
     gtk_widget_queue_draw_area(GTK_WIDGET(display),
-                               (d->mouse_guest_x + d->mx - d->mouse_hotspot.x) * sx,
-                               (d->mouse_guest_y + d->my - d->mouse_hotspot.y) * sy,
+                               (d->mouse_guest_x + d->mx - d->mouse_hotspot.x - d->area.x) * sx,
+                               (d->mouse_guest_y + d->my - d->mouse_hotspot.y - d->area.y) * sy,
                                gdk_pixbuf_get_width(d->mouse_pixbuf) * sx,
                                gdk_pixbuf_get_height(d->mouse_pixbuf) * sy);
 }
@@ -1749,50 +1926,6 @@ static void cursor_reset(SpiceCursorChannel *channel, gpointer data)
     gdk_window_set_cursor(window, NULL);
 }
 
-static void disconnect_main(SpiceDisplay *display)
-{
-    SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
-
-    if (d->main == NULL)
-        return;
-    g_signal_handlers_disconnect_by_func(d->main, G_CALLBACK(update_mouse_mode),
-                                         display);
-    d->main = NULL;
-}
-
-static void disconnect_display(SpiceDisplay *display)
-{
-    SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
-
-    if (d->display == NULL)
-        return;
-    g_signal_handlers_disconnect_by_func(d->display, G_CALLBACK(primary_create),
-                                         display);
-    g_signal_handlers_disconnect_by_func(d->display, G_CALLBACK(primary_destroy),
-                                         display);
-    g_signal_handlers_disconnect_by_func(d->display, G_CALLBACK(invalidate),
-                                         display);
-    primary_destroy(d->display, display);
-    d->display = NULL;
-}
-
-static void disconnect_cursor(SpiceDisplay *display)
-{
-    SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
-
-    if (d->cursor == NULL)
-        return;
-    g_signal_handlers_disconnect_by_func(d->cursor, G_CALLBACK(cursor_set),
-                                         display);
-    g_signal_handlers_disconnect_by_func(d->cursor, G_CALLBACK(cursor_move),
-                                         display);
-    g_signal_handlers_disconnect_by_func(d->cursor, G_CALLBACK(cursor_hide),
-                                         display);
-    g_signal_handlers_disconnect_by_func(d->cursor, G_CALLBACK(cursor_reset),
-                                         display);
-    d->cursor = NULL;
-}
-
 static void channel_new(SpiceSession *s, SpiceChannel *channel, gpointer data)
 {
     SpiceDisplay *display = data;
@@ -1802,25 +1935,34 @@ static void channel_new(SpiceSession *s, SpiceChannel *channel, gpointer data)
     g_object_get(channel, "channel-id", &id, NULL);
     if (SPICE_IS_MAIN_CHANNEL(channel)) {
         d->main = SPICE_MAIN_CHANNEL(channel);
-        g_signal_connect(channel, "main-mouse-update",
-                         G_CALLBACK(update_mouse_mode), display);
+        spice_g_signal_connect_object(channel, "main-mouse-update",
+                                      G_CALLBACK(update_mouse_mode), display, 0);
         update_mouse_mode(channel, display);
         return;
     }
 
     if (SPICE_IS_DISPLAY_CHANNEL(channel)) {
+        SpiceDisplayPrimary primary;
         if (id != d->channel_id)
             return;
         d->display = channel;
-        g_signal_connect(channel, "display-primary-create",
-                         G_CALLBACK(primary_create), display);
-        g_signal_connect(channel, "display-primary-destroy",
-                         G_CALLBACK(primary_destroy), display);
-        g_signal_connect(channel, "display-invalidate",
-                         G_CALLBACK(invalidate), display);
-        g_signal_connect(channel, "display-mark",
-                         G_CALLBACK(mark), display);
+        spice_g_signal_connect_object(channel, "display-primary-create",
+                                      G_CALLBACK(primary_create), display, 0);
+        spice_g_signal_connect_object(channel, "display-primary-destroy",
+                                      G_CALLBACK(primary_destroy), display, 0);
+        spice_g_signal_connect_object(channel, "display-invalidate",
+                                      G_CALLBACK(invalidate), display, 0);
+        spice_g_signal_connect_object(channel, "display-mark",
+                                      G_CALLBACK(mark), display, G_CONNECT_AFTER | G_CONNECT_SWAPPED);
+        spice_g_signal_connect_object(channel, "notify::monitors",
+                                      G_CALLBACK(update_monitor_area), display, G_CONNECT_AFTER | G_CONNECT_SWAPPED);
+        if (spice_display_get_primary(channel, 0, &primary)) {
+            primary_create(channel, primary.format, primary.width, primary.height,
+                           primary.stride, primary.shmid, primary.data, display);
+            mark(display, primary.marked);
+        }
         spice_channel_connect(channel);
+        spice_main_set_display_enabled(d->main, get_display_id(display), TRUE);
         return;
     }
 
@@ -1828,14 +1970,14 @@ static void channel_new(SpiceSession *s, SpiceChannel *channel, gpointer data)
         if (id != d->channel_id)
             return;
         d->cursor = SPICE_CURSOR_CHANNEL(channel);
-        g_signal_connect(channel, "cursor-set",
-                         G_CALLBACK(cursor_set), display);
-        g_signal_connect(channel, "cursor-move",
-                         G_CALLBACK(cursor_move), display);
-        g_signal_connect(channel, "cursor-hide",
-                         G_CALLBACK(cursor_hide), display);
-        g_signal_connect(channel, "cursor-reset",
-                         G_CALLBACK(cursor_reset), display);
+        spice_g_signal_connect_object(channel, "cursor-set",
+                                      G_CALLBACK(cursor_set), display, 0);
+        spice_g_signal_connect_object(channel, "cursor-move",
+                                      G_CALLBACK(cursor_move), display, 0);
+        spice_g_signal_connect_object(channel, "cursor-hide",
+                                      G_CALLBACK(cursor_hide), display, 0);
+        spice_g_signal_connect_object(channel, "cursor-reset",
+                                      G_CALLBACK(cursor_reset), display, 0);
         spice_channel_connect(channel);
         return;
     }
@@ -1868,21 +2010,22 @@ static void channel_destroy(SpiceSession *s, SpiceChannel *channel, gpointer dat
     SPICE_DEBUG("channel_destroy %d", id);
 
     if (SPICE_IS_MAIN_CHANNEL(channel)) {
-        disconnect_main(display);
+        d->main = NULL;
         return;
     }
 
     if (SPICE_IS_DISPLAY_CHANNEL(channel)) {
         if (id != d->channel_id)
             return;
-        disconnect_display(display);
+        primary_destroy(d->display, display);
+        d->display = NULL;
         return;
     }
 
     if (SPICE_IS_CURSOR_CHANNEL(channel)) {
         if (id != d->channel_id)
             return;
-        disconnect_cursor(display);
+        d->cursor = NULL;
         return;
     }
 
@@ -1904,7 +2047,7 @@ static void channel_destroy(SpiceSession *s, SpiceChannel *channel, gpointer dat
 /**
  * spice_display_new:
  * @session: a #SpiceSession
- * @id: the display channel ID to associate with #SpiceDisplay
+ * @channel_id: the display channel ID to associate with #SpiceDisplay
  *
  * Returns: a new #SpiceDisplay widget.
  **/
@@ -1914,6 +2057,24 @@ SpiceDisplay *spice_display_new(SpiceSession *session, int id)
                         "channel-id", id, NULL);
 }
 
+/**
+ * spice_display_new_with_monitor:
+ * @session: a #SpiceSession
+ * @channel_id: the display channel ID to associate with #SpiceDisplay
+ * @monitor_id: the monitor id within the display channel
+ *
+ * Since: 0.13
+ * Returns: a new #SpiceDisplay widget.
+ **/
+SpiceDisplay* spice_display_new_with_monitor(SpiceSession *session, gint channel_id, gint monitor_id)
+{
+    return g_object_new(SPICE_TYPE_DISPLAY,
+                        "session", session,
+                        "channel-id", channel_id,
+                        "monitor-id", monitor_id,
+                        NULL);
+}
+
 /**
  * spice_display_mouse_ungrab:
  * @display:
@@ -1976,16 +2137,14 @@ GdkPixbuf *spice_display_get_pixbuf(SpiceDisplay *display)
 
     g_return_val_if_fail(d != NULL, NULL);
     /* TODO: ensure d->data has been exposed? */
+    g_return_val_if_fail(d->data != NULL, NULL);
 
-    data = g_malloc(d->width * d->height * 3);
+    data = g_malloc(d->area.width * d->area.height * 3);
     src = d->data;
     dest = data;
 
-    if (src == NULL || dest == NULL)
-        return NULL;
-
-    for (y = 0; y < d->height; ++y) {
-        for (x = 0; x < d->width; ++x) {
+    for (y = d->area.y; y < d->area.height; ++y) {
+        for (x = d->area.x; x < d->area.width; ++x) {
           dest[0] = src[x * 4 + 2];
           dest[1] = src[x * 4 + 1];
           dest[2] = src[x * 4 + 0];
@@ -1995,7 +2154,7 @@ GdkPixbuf *spice_display_get_pixbuf(SpiceDisplay *display)
     }
 
     pixbuf = gdk_pixbuf_new_from_data(data, GDK_COLORSPACE_RGB, false,
-                                      8, d->width, d->height, d->width * 3,
+                                      8, d->area.width, d->area.height, d->area.width * 3,
                                       (GdkPixbufDestroyNotify)g_free, NULL);
     return pixbuf;
 }
